#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Wed 05 May 2010 01:23:35 PM CST

''' date, zone, and so '''

import os
import time

from system import _all
from system import POPEN
from system import NASError
from system import get_name_value
from system import set_name_value

from service import startservice
from service import stopservice
from service import enableservice

from login import updatetime

NTPCONF = '/etc/ntp.conf'
ZONETAB = '/usr/share/zoneinfo/zone.tab'

def getzone(**kw):
    ''' current time zone '''
    return {'zone': time.strftime('%Z')}

def getdate(**kw):
    ''' current server date '''
    date = time.strftime('%F %H:%M:%S %Z')
    return {'date': date}

def getntp(**kw):
    ''' current ntp server '''
    names = {'server': []}
    get_name_value(NTPCONF, names, ['#'], ' ')
    return names

def setzone(zone=None, **kw):
    ''' set time zone '''
    if zone is None:
        raise NASError('Bad Request')

    zones = _parsezones()
    if zone not in zones:
        raise NASError('Invalid Zone', zone)

    path = '/usr/share/zoneinfo/%s' % zone
    localtime = '/etc/localtime'
    if os.path.lexists(localtime):
        os.unlink(localtime)
    os.symlink(path, localtime)
    return {'date': 'update time zone'}

def _systohc(session=None, **kw):
    ''' sync system time to hardware date '''

    if session is None:
        raise NASError('Bad Request')

    session = str(session)

    (error, pipe) = POPEN(['hwclock', '-w'])
    status = pipe.read()
    pipe.close()
    if error == 0:
        updatetime(session, newdate=True)
        return {'date': 'update system time'}

    raise NASError('systohc', status)

def setdate(session=None, date=None, **kw):
    ''' set date, the format is unspecified '''

    if not _all(session, date):
        raise NASError('Bad Request')

    (error, pipe) = POPEN(['date', '-s', str(date)])
    status = pipe.read()
    pipe.close()

    if error == 0:
        return _systohc(session)

    raise NASError('setdate', status)

def setntp(ntp=None, **kw):
    ''' set ntp server '''

    if ntp is None:
        raise NASError('Bad Request')

    if isinstance(ntp, list):
        ntp = [str(x) for x in ntp]
    else:
        ntp = [str(ntp)]

    names = {'server': ntp}
    set_name_value(NTPCONF, names, ['#'], ' ')
    return {'date': 'update ntp server'}

def getzones(**kw):
    ''' get time zones '''
    zones = _parsezones()
    zones.sort()
    return {'zone': zones}

def syncdate(session=None, **kw):
    ''' sync date with ntp server '''
    if session is None:
        raise NASError('Bad Request')

    stopservice('ntpdate')
    enableservice('ntpdate')
    startservice('ntpdate')
    return _systohc(session)

def _parsezones():
    ''' internal function for parse zonetab '''
    zones = []
    _file = open(ZONETAB)
    while True:
        line = _file.readline()
        if not line:
            break
        if line.startswith('#'):
            continue
        zone = line.split('\t')
        if len(zone) > 2:
            zones.append(zone[2].strip())
    _file.close()
    return zones

__all__ = ['getzone', 'getzones', 'setzone',
           'getdate', 'setdate',
           'getntp', 'setntp' 'syncdate']

# vim: set sta sw=4 et:
