#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Mon 19 Jul 2010 03:03:42 PM CST

''' facebook upload photo '''

import os
import httplib
import socket

import mimetypes
mimetypes.init()

HOST = 'graph.facebook.com'
PORT = 443

from system import NASError

import json
import hashlib
from cStringIO import StringIO
from system import _all
from system import getrealpath

class FacebookError(NASError):
    def __init__(self, detail):
        raise NASError('Facebook Error', detail)

def _getdata(path, headers=None):
    connection = httplib.HTTPSConnection(HOST, PORT)
    connection.putrequest('GET', path)
    if headers is not None:
        for (name, value) in headers:
            connection.putheader(name, value)
    try:
        connection.endheaders()
    except socket.error as error:
        raise NASError('Proxy Error', error.strerror)
    return _getresult(connection)

def _postdata(path, body, headers=None):
    connection = httplib.HTTPSConnection(HOST, PORT)
    connection.putrequest('POST', path)

    has_type = False
    has_length = False
    if headers is not None:
        for (name, value) in headers:
            if name.lower() == 'content-length':
                has_length = True
            if name.lower() == 'content-type':
                has_type = True
            connection.putheader(name, value)

    if not has_type:
        connection.putheader('Content-Type',
                'application/x-www-form-urlencoded')

    if not has_length:
        connection.putheader('Content-Length', str(len(body)))

    try:
        connection.endheaders()
    except socket.error as error:
        raise FacebookError(error.strerror)
    connection.send(body)
    if hasattr(body, 'close'):
        body.close()

    return _getresult(connection)

def _getresult(connection):
    try:
        response = connection.getresponse()
    except httplib.HTTPException as error:
        connection.close()
        raise NASError('Proxy Error', error.reason)

    content = response.read()
    connection.close()
    if response.status != 200:
        raise FacebookError('Invalid Status')

    try:
        result = json.loads(content)
        if result.get('error_code') is not None:
            raise FacebookError(result.get('error_code'))
        return result
    except ValueError:
        raise FacebookError('Invalid Response')

def _multipart(query):
    boundary = hashlib.sha1(os.urandom(20)).hexdigest()

    body = StringIO()
    for item in query.keys():
        value = query.get(item)
        if not value:
            continue
        body.write('--%s\r\n' % boundary)
        if isinstance(value, str):
            body.write('Content-Disposition: form-data; ')
            body.write('name="%s"\r\n' % item)
            body.write('\r\n')
            body.write(value)
            body.write('\r\n')
        elif hasattr(value, 'read'):
            content_type = mimetypes.guess_type(value.name)[0]
            body.write('Content-Disposition: form-data; ')
            body.write('name="%s"; ' % item)
            body.write('filename="%s"\r\n' % value.name)
            body.write('Content-Type: %s\r\n' % content_type)
            body.write('\r\n')
            body.write(value.read())
            body.write('\r\n')
            value.close()
    body.write('--%s--\r\n' % boundary)
    headers = []
    headers.append(('Content-Type',
                    'multipart/form-data; boundary=%s' % boundary))
    headers.append(('Content-Length', str(body.tell())))
    body.seek(0, 0)
    return (headers, body)

def uploadphoto(access_token=None, path=None, login=None,
        albumid='me', message=None, **kw):
    if not _all(access_token, path):
        raise NASError('Bad Request')

    realpath = getrealpath(path, login, False)
    if not realpath:
        raise NASError('Not Found')

    content_type = mimetypes.guess_type(realpath)[0]
    if not content_type.startswith('image'):
        raise NASError('Not Image')

    path = '/%s/photos' % albumid
    query = {'access_token': access_token,
             os.path.basename(realpath): open(realpath)}
    if message is not None:
        query['message'] = message
    headers, body = _multipart(query)
    data = _postdata(path, body, headers)
    if 'id' in data:
        path = '/%s?access_token=%s' % (data['id'], access_token)
        return _getdata(path)
    return data

__all__ = ['uploadphoto']

# vim: set sta sw=4 et:
