#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Wed 14 Apr 2010 08:47:29 AM CST

''' Infomation and Regester '''

import os
import time
import urllib
import urllib2
import hashlib
import fcntl

from system import _all
from system import HOME
from system import NASError
from system import get_name_value
from system import set_name_value
from system import getvendor
from system import _canwrite
from system import _mkdir
from system import backup, unbackup
from system import lock, unlock

from xml.dom.minidom import Node
from xml.dom.minidom import parseString

from service import stopservice
from service import startservice
from service import enableservice

from network import gethwaddr, getipaddr

DEVICEVERSION = '/etc/nas/version'
DEVICEINFO = '/etc/nas/device'
SAMBACONF = '/etc/nas/smb.conf'

def _openvpn(environ):
    from network import getdev
    if getdev(environ.get('SERVER_ADDR')) in ('tun', 'tap'):
        return 'yes'
    return 'no'

def getregister(**kw):
    info = {'register': '', 'upnpaddr': ''}
    info['openvpn'] = _openvpn(kw.get('environ', {}))
    info['ipaddr'] = getipaddr()
    get_name_value(DEVICEINFO, info)
    return info

def getinfo(**kw):
    info = {'register': '', 'workgroup': '', 'hardware': '', 'software': '', 'upnpaddr': ''}
    get_name_value(DEVICEVERSION, info)
    get_name_value(DEVICEINFO, info)
    get_name_value(SAMBACONF, info)

    info['hostname'] = os.uname()[1]
    info['openvpn'] = _openvpn(kw.get('environ', {}))
    info['ipaddr'] = getipaddr()
    info['hwaddr'] = gethwaddr()
    info['date'] = time.strftime('%F %H:%M:%S %Z')
    info['gui'] = getvendor()

    storage = {}
    vfs = os.statvfs(HOME)
    storage['blocks'] = str(vfs.f_blocks * vfs.f_bsize)
    storage['used'] = str((vfs.f_blocks - vfs.f_bavail) * vfs.f_bsize)
    storage['available'] = str(vfs.f_bavail * vfs.f_bsize)
    storage['usedpercent'] = str(100 - 100 * vfs.f_bavail / vfs.f_blocks)
    storage['writable'] = _canwrite() and 'yes' or 'no'
    info['storage'] = storage

    if not info['register']:
        info['register'] = info['hwaddr'].replace(':', '')

    return {'info': info}

def _setcert(host, port, param):

    url = 'http://%s:%s/gui/register_device' % (host, port)

    request = urllib2.Request(url, urllib.urlencode(param))
    try:
        data = urllib2.urlopen(request)
    except urllib2.URLError as error:
        raise NASError('Register Fail', 'Connection Error: %s' % error.reason)
    except Exception as error:
        raise NASError('Register Fail', error.message)

    content = data.read()
    data.close()

    xml = parseString(content)

    _ca = _get_node_value(xml, 'CA')
    _cert = _get_node_value(xml, 'CERT')
    _key = _get_node_value(xml, 'KEY')
    # _conf = _get_node_value(xml, 'CONF')

    CONF = ['client',
            'dev tun',
            'proto udp',
            'remote %s 1194' % host,
            'resolv-retry infinite',
            'nobind',
            'persist-key',
            'persist-tun',
            'shaper 57344',
            'ca ca.crt',
            'cert client.crt',
            'key client.key',
            'comp-lzo',
            'verb 3',]

    _conf = '\n'.join(CONF)

    if not _all(_ca, _cert, _key, _conf):
        raise NASError('Register Fail', 'Invalid Response: %s' % content)

    if not os.path.isdir('/etc/openvpn'):
        if os.path.lexists('/etc/openvpn'):
            os.unlink('/etc/openvpn')
        _mkdir('/etc/openvpn')

    for x in ['ca.crt', 'client.crt', 'client.key', 'client.conf']:
        backup('%s/%s' % ('/etc/openvpn', x))

    def write(name, data):
        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC | os.O_SYNC
        pipe = os.open('%s.tmp' % name, flags, 0644)
        fcntl.lockf(pipe, fcntl.LOCK_EX)
        retcode = os.write(pipe, data)
        os.fsync(pipe)
        fcntl.lockf(pipe, fcntl.LOCK_UN)
        os.close(pipe)
        if retcode != len(data):
            raise NASError('write data error', name)

    write('/etc/openvpn/ca.crt', _ca)
    write('/etc/openvpn/client.crt', _cert)
    write('/etc/openvpn/client.key', _key)
    write('/etc/openvpn/client.conf', _conf)

    os.rename('/etc/openvpn/ca.crt.tmp', '/etc/openvpn/ca.crt')
    os.rename('/etc/openvpn/client.crt.tmp', '/etc/openvpn/client.crt')
    os.rename('/etc/openvpn/client.key.tmp', '/etc/openvpn/client.key')
    os.rename('/etc/openvpn/client.conf.tmp', '/etc/openvpn/client.conf')

    for x in ['ca.crt', 'client.crt', 'client.key', 'client.conf']:
        unbackup('%s/%s' % ('/etc/openvpn', x))

cache = {}
PLX7821 = '-fast-' in os.uname()[2]
def getseckey():
    if cache.get('seckey'):
        return cache.get('seckey')

    if not PLX7821:
        return ''

    NEW = 0xa40000
    PAGE = 0x800
    mtd1 = "/dev/mtdblock1"
    seckey = ''
    seckey1 = seckey2 = None
    pipe = None
    try:
        pipe = os.open(mtd1, os.O_RDONLY)
        os.lseek(pipe, NEW - PAGE * 2 + 6, os.SEEK_SET)
        seckey1 = os.read(pipe, 16)
        os.lseek(pipe, NEW - PAGE + 6, os.SEEK_SET)
        seckey2 = os.read(pipe, 16)
    except os.error:
        pass
    finally:
        if pipe is not None:
            os.close(pipe)

    if seckey1 is not None and seckey1 == seckey2:
        seckey = seckey1
        cache['seckey'] = seckey

    return seckey

def register(name=None, host=None, port='80', **kw):
    hwaddr = gethwaddr().replace(':', '')
    if not hwaddr:
        raise NASError("No MAC Address")

    if name is None:
        name = hwaddr
    else:
        name = str(name)

    if host is None or not str(host):
        host = 'isharing.workssys.com'
    else:
        host = str(host)

    seckey = getseckey()
    param = {'Key': hwaddr,
             'Device_Name': name,
             }
    if seckey:
        sec_key = '%s:%s' % (name, seckey)
        param['sec_key'] = hashlib.md5(sec_key).hexdigest()

    pipe = lock('/var/tmp/openvpn')
    _setcert(host, port, param)
    unlock(pipe)

    stopservice('openvpn')

    info = {'register': name}
    set_name_value(DEVICEINFO, info)

    enableservice('openvpn')
    startservice('openvpn')
    return {'register': 'done'}

def _get_node_value(xml, tag):
    nodes = xml.getElementsByTagName(tag)
    if nodes.length != 1:
        return None
    node = nodes.item(0)
    if not node.hasChildNodes():
        return None
    nodes = node.childNodes
    if nodes.length != 1:
        return None
    node = nodes.item(0)
    if not isinstance(node, Node):
        return None
    return node.nodeValue

__all__ = ['getregister', 'getinfo', 'register']

# vim: set sta sw=4 et:
