#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Mon 10 May 2010 05:40:24 PM CST

''' Login, Decrypt Password '''

RSABIT = 512
TIMEOUT = 60 * 60 * 24

import time
from system import NASError
from system import sqlite
from system import SESSIONDB
from system import _all
from system import makesure_nasdir
from system import getunicode
from system import decrypt
from hashlib import sha1
from rsa import generate
from users import _checkhomedir
from users import _checkpass
from info import getinfo

def genhash(**kw):
    ''' Generate Hash and RSA Key
        return: hash, modules, public
    '''
    def hexstr(x=0):
        y = unicode(hex(x))
        if y.endswith('L'):
            return y[2:-1]
        else:
            return y[2:]

    session = unicode(sha1(str(time.time())).hexdigest())

    public, modulus, private = [hexstr(x) for x in generate(RSABIT)]

    con = sqlite.connect(SESSIONDB)
    cur = con.cursor()
    cur.execute(''' DELETE FROM session WHERE hash = ?
                ''', (session,))
    cur.execute(''' DELETE FROM session WHERE time < ?
                ''', (time.time() - TIMEOUT,))
    cur.execute(''' INSERT INTO session (hash, modulus, public, private, time)
                    VALUES (?, ?, ?, ?, ?)
                ''', (session, modulus, public, private, time.time()))
    try:
        cur.execute(''' VACUUM session ''')
    except:
        pass
    con.commit()
    con.close()

    makesure_nasdir()
    return {'hash': session.encode('utf8'),
            'modulus': modulus.encode('utf8'),
            'public': public.encode('utf8')}

# rewrite gethash
def gethash(session=None, **kw):
    ''' Get RSA Key for Hash
        param : hash ->
        return: modulus, public
    '''
    if session is None:
        raise NASError('Bad Request')

    session = getunicode(session)

    con = sqlite.connect(SESSIONDB)
    cur = con.cursor()
    cur.execute(''' SELECT modulus, public FROM session WHERE hash = ?
                ''', (session,))
    result = cur.fetchone()
    con.close()

    if not result:
        raise NASError('Bad Request')

    return {'modulus': result[0].encode('utf8'),
            'public': result[1].encode('utf8')}

def updatetime(session, username=None, newdate=False, **kw):
    ''' Update time
        param : session
        return: (False, reason), (True, user)
    '''
    if session is None:
        raise NASError('Bad Request')

    session = getunicode(session)

    con = sqlite.connect(SESSIONDB)
    cur = con.cursor()

    cur.execute(''' SELECT id, time, user FROM session WHERE hash = ?
                ''', (session,))

    result = cur.fetchone()
    if not result:
        con.close()
        raise NASError('Bad Request')

    if not newdate and time.time() - result[1] > TIMEOUT:
        cur.execute(''' DELETE FROM session WHERE id = ?
                    ''', result[:1])
        con.commit()
        con.close()
        raise NASError('Timeout')

    cur.execute(''' DELETE FROM session WHERE time < ?
                ''', (time.time() - TIMEOUT,))

    if username is not None:
        user = str(username).decode('utf8', 'ignore')
        cur.execute(''' UPDATE session SET time = ?, user = ?
                        WHERE id = ?
                    ''', (time.time(), user, result[0]))
    else:
        cur.execute(''' UPDATE session SET time = ?
                        WHERE id = ?
                    ''', (time.time(), result[0]))
        user = result[2]

    try:
        cur.execute(''' VACUUM session ''')
    except:
        pass

    con.commit()
    con.close()

    if not user:
        raise NASError('Not Login')

    _checkhomedir(user.encode('utf8'))
    return user.encode('utf8')

def logout(session=None, **kw):
    ''' Delete Hash
        param : session
        return: False, True
    '''
    if session is None:
        raise NASError('Bad Request')

    session = getunicode(session)

    con = sqlite.connect(SESSIONDB)
    cur = con.cursor()
    cur.execute(''' DELETE FROM session WHERE hash = ?
                ''', (session,))
    con.commit()
    con.close()
    return {'logout': session.encode('utf8')}

def login(session=None, username=None, password=None, **kw):
    ''' user -> username, pass->password
        return: info in getinfo module
    '''
    if username is None:
        username = kw.get('user')
    if password is None:
        password = kw.get('pass')

    if not _all(session, username, password):
        raise NASError('Bad Request')

    username = str(username)
    password = decrypt(session, password)

    if not _checkpass(username, password):
        raise NASError('Wrong Username or Password')

    updatetime(session, username)

    return getinfo()

__all__ = ['genhash', 'gethash', 'updatetime', 'logout', 'login']

# vim: set sta sw=4 et:
