#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Tue 03 Aug 2010 10:13:52 AM CST

''' NFS Service '''
EXPORTS = '/etc/exports'

import re
from service import *
from system import POPEN
from system import NASError
from system import getrealpath
from system import backup, unbackup
from cStringIO import StringIO

import os
import stat
import fcntl

def getexports(**kw):
    exports = {}
    fp = open(EXPORTS)
    while True:
        line = fp.readline()
        if not line:
            break
        line = line.strip()
        if line[:1] == '#':
            continue
        lists = line.split()
        if len(lists) < 2:
            continue
        name = lists[0]
        directory = {'name': name, 'machine': []}
        for option in lists[1:]:
            find = re.findall('^(.+)\((.+)\)$', option)
            if len(find) != 1:
                continue
            host = find[0][0]
            options = find[0][1]
            directory['machine'].append({'host': host, 'options': options})
        if 'directory' not in exports:
            exports['directory'] = [directory]
        else:
            exports['directory'].append(directory)
    fp.close()
    (unused, pipe) = POPEN('exportfs -r')
    warning = pipe.read()
    pipe.close()
    exports['warning'] = warning
    exports['enable'] = servicestatus('nfs')
    exports['running'] = servicerunning('rpc.mountd')

    return {'nfs': exports}

def setexports(login=None, enable=None, directory=None, **kw):
    if isinstance(directory, list):
        raise NASError('Bad Request')

    if directory is not None:
        path = getrealpath(directory, login, True, system=True)
        host = kw.get('host')
        if not path:
            if directory[:4] == '/dav':
                raise NASError('path not exists')
            else:
                path = directory
                host = None

        for escape in ['\040', '\011', '\012', '\134']:
            if escape in path:
                raise NASError('Unsupported Directory %s' % directory,
                               '" ", "\\t", "\\n" and "\\\\" are unsupported.')
        _set_exports(path, host, kw.get('options'))

    if str(enable) == 'yes':
        enableservice('nfs')
        restartservice('nfs', 'rpc.mountd')
    elif enable is not None:
        disableservice('nfs')
        stopservice('nfs')

    return getexports()

def _set_exports(directory, host, options):

    changed, data = _merge_exports(directory, host, options)

    content = data.getvalue()
    data.close()

    if changed:
        backup(EXPORTS)
        flag = os.O_WRONLY | os.O_CREAT | os.O_TRUNC | os.O_SYNC
        mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
        fd = os.open('%s.tmp' % EXPORTS, flag, mode)
        fcntl.lockf(fd, fcntl.LOCK_EX)
        ret = os.write(fd, content)
        os.fsync(fd)
        fcntl.lockf(fd, fcntl.LOCK_UN)
        os.close(fd)
        if ret != len(content):
            raise NASError('write data error', EXPORTS)
        os.rename('%s.tmp' % EXPORTS, EXPORTS)
        os.chown(EXPORTS, 0, 0)
        POPEN('exportfs -r')[1].close()
        unbackup(EXPORTS)


def _merge_option(newline, lists, host, options, info):
    changed = False
    for option in lists:
        finds = re.findall('^(.+)\((.+)\)$', option)
        if len(finds) == 1:
            _host = finds[0][0]
            _option = finds[0][1]
            if _host in host:
                index = host.index(_host)
                try:
                    _option = options[index]
                except IndexError:
                    continue
                host.pop(index)
                options.pop(index)
                changed = True
                newoption = '%s(%s)' % (_host, _genoption(info, _option))
                newline.append(newoption)
    return changed

def _merge_exports(directory, host, options):

    data = StringIO()
    merged = changed = False

    try:
        info = os.stat(directory)
        infostr = ''
    except os.error as error:
        info = None
        infostr = error.strerror

    _file = open(EXPORTS)
    while True:
        line = _file.readline()
        if not line:
            break
        line = line.strip()
        if line[:1] == '#':
            continue
        lists = line.split()
        if len(lists) < 2:
            continue
        if lists.pop(0) != directory:
            data.write(line + '\n')
            continue

        merged = True

        if host is None:
            changed = True
            continue

        if not isinstance(host, list):
            host = [host]
        if not isinstance(options, list):
            options = [options]

        if not info:
            raise NASError('Invalid Directory', infostr)

        newline = [directory]
        if _merge_option(newline, lists, host, options, info):
            changed = True

        for machine in zip(host, options):
            changed = True
            newoption = '%s(%s)' % (machine[0], _genoption(info, machine[1]))
            newline.append(newoption)

        if changed:
            data.write(' '.join(newline) + '\n')

    _file.close()

    # add new record
    if not merged:
        if not info:
            raise NASError('Invalid Directory', infostr)
        newline = [directory]
        if not isinstance(host, list):
            host = [host]
        if not isinstance(options, list):
            options = [options]
        for machine in zip(host, options):
            newoption = '%s(%s)' % (machine[0], _genoption(info, machine[1]))
            newline.append(newoption)
        changed = True
        data.write(' '.join(newline) + '\n')

    return changed, data

def _genoption(info, option):
    if 'rw' in option:
        rw = 'rw'
    else:
        rw = 'ro'
    nfs_option = ''
    nfs_option += 'async,no_subtree_check,'
    nfs_option += '%s,all_squash,' % rw
    nfs_option += 'anonuid=%s,anongid=%s' % (info.st_uid, info.st_gid)
    return nfs_option

__all__ = ['getexports', 'setexports']

# vim: set sta sw=4 et:
