#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Sun 27 Jun 2010 12:35:07 PM CST

''' Spin Down '''

import os
import re

from system import POPEN
from system import NASError
from system import get_name_value
from system import set_name_value

def _resolvetime(seconds):
    assert isinstance(seconds, int)
    def _divmod(a, b):
        c, d = divmod(a, b)
        e = d and c + 1 or c
        return e
    # 5s ~ 20m
    if seconds <= 20 * 60:
        return _divmod(seconds, 5)
    # 21m is special
    elif seconds <= 21 * 60:
        return 252
    # 21m + 15s is special, too
    elif seconds <= 21 * 60 + 15:
        return 255
    # 30m ~ 5.5h
    elif seconds <= 19800:
        return _divmod(seconds, 1800) + 240

def _resolvetimeout(value):
    assert isinstance(value, int)
    assert value != 253, 'not support vendor specific timeout'
    assert value != 254
    assert value < 256
    if value <= 240:
        return value * 5
    elif value <= 251:
        return 1800 * (value - 240)
    elif value == 252:
        return 1800 * 22
    elif value == 255:
        return 1800 * 22 + 15

def getspindown(**kw):
    data = {'TIMEOUT': '',
            'SECONDS': ''}
    get_name_value('/etc/nas/spindown', data)
    return {'timeout': data['TIMEOUT'],
            'seconds': data['SECONDS']}

def _spindown(path):
    disk = path.rstrip('0123456789')
    if os.path.exists(disk):
        data = getspindown()
        error, result = POPEN('hdparm -S%s %s' % (data['timeout'], disk))
        result.close()

def _getseconds(timeout):
    unit = 1
    if timeout.endswith('s'):
        time = timeout[:-1]
    elif timeout.endswith('m'):
        time = timeout[:-1]
        unit = 60
    elif timeout.endswith('h'):
        time = timeout[:-1]
        unit = 3600
    else:
        time = timeout[:]

    try:
        time = float(time)
    except ValueError:
        raise NASError('Bad Request')

    return int(time * unit)

def spindown(disks=None, timeout=None, **kw):
    if timeout is None:
        raise NASError('Bad Request')

    seconds = _getseconds(timeout)
    timeout = _resolvetime(seconds)

    if disks is None:
        disks = []
        for path in os.listdir('/dev'):
            if re.match(r'^[s|h]d[a-z]$', path):
                disks.append(os.path.join('/dev', path))
    elif not isinstance(disks, list):
        disks = [str(disks)]
    else:
        disks = [str(x) for x in disks]

    if not disks:
        raise NASError('No SpinDown Disks')

    statuses = []
    seconds = _resolvetimeout(timeout)

    data = {'TIMEOUT': timeout,
            'SECONDS': seconds}
    set_name_value('/etc/nas/spindown', data)

    for disk in disks:
        (error, result) = POPEN('hdparm -S%s %s' % (timeout, disk))
        status = result.read()
        result.close()
        if error == 0:
            statuses.append({'spindown': {'disk': disk, 'result': status}})
        else:
            statuses.append({'error': {'reason': disk, 'detail': status}})
    return {'spindown': statuses}

__all__ = ['getspindown', 'spindown']

# vim: set sta sw=4 et:
