#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Thu 01 Jul 2010 08:49:28 PM CST

import os
import re
# import sys
import stat
import time
import errno
import fcntl
import urllib
import urlparse
import shutil
import Image
import backup
from xml.dom.minidom import parseString

from system import url2path
from system import fixperm
from system import writexml
from system import REALFIRMWARE
from system import NASError
from system import StatusError
from system import PathError
from system import getchildren
from system import BACKUP, REALRESTORE

from exif import getexif
from firmware import uncompress

import pwd
import convert

from uuid import uuid4
from cStringIO import StringIO

ROOT = 'root'
HOME = 'home'
PUBLIC = 'public'
SHARE = 'share'
DEVICE = 'device'
HOMES = 'homes'

TZ = '%Y-%m-%dT%H:%M:%SZ'
RFC1123 = '%a, %d %b %Y %H:%M:%S GMT'
DTNS = 'xmlns:dt="urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/"'

MODE = 00664

import magic
mime = magic.open(magic.MAGIC_MIME)
mime.load()
import mimetypes
mimetypes.init()
def _getmimetype(path):
    mimetype = mime.file(path)
    if not mimetype or 'charset=' not in mimetype:
        mimetype = 'application/octet-stream'
    else:
        mimetype = mimetype.split(';')[0]
    if mimetype[:4] == 'text':
        suffix = mimetypes.guess_type(path)[0]
        if suffix: mimetype = suffix
    elif mimetype == 'application/x-empty':
        mimetype = 'text/plain'
    return mimetype

def _getdisposition(path, **kw):
    filename = os.path.split(path)[1]
    # quote is need for ie
    if kw.get('quote', '').lower() == 'true':
        filename = urllib.quote(filename)
    disposition = 'attachments; filename="%s"' % filename
    return disposition

def _getetag(sb):
    return '"%x-%x-%x"' % (sb.st_ino, sb.st_mtime, sb.st_size)

def _getmtime(sb):
    return time.strftime(RFC1123, time.gmtime(sb.st_mtime))

def _modified(old, sb):
    if not old:
        return None

    try:
        oldtime = time.strptime(old, RFC1123)
    except ValueError:
        return None

    if oldtime >= time.gmtime(sb.st_mtime):
        return False

    return True

def _matched(old, sb):
    if not old:
        return None
    etag = _getetag(sb)
    for tag in old.split(','):
        if '*' in tag:
            return True
        if tag.strip() == etag:
            return True
    return False

def checkif(sb, environ):
    if _modified(environ.get('If-Modified-Since', ''), sb) is False:
        raise StatusError(304)
    if _modified(environ.get('If-Unmodified-Since', ''), sb) is True:
        raise StatusError(412)
    if _matched(environ.get('If-Matched', ''), sb) is False:
        raise StatusError(412)
    if _matched(environ.get('If-None-Matched', ''), sb) is True:
        if environ.get('REQUEST_METHOD', '') in ('GET', 'HEAD'):
            raise StatusError(304)
        raise StatusError(412)

# Always use X-Sendfile for file downloading
PLX7821 = False
# PLX7821 = '-fast-' in os.uname()[2]

def _readfile(path, start=None, end=None):
    body = []
    try:
        _file = open(path, 'rb')
    except IOError:
        return body
    if start is None or end is None:
        body.append(_file.read())
    else:
        _file.seek(start)
        body.append(_file.read(end - start + 1))
    _file.close()
    return body

def _xsendfile(path, environ, size, headers, xsend=True):
    if isinstance(path, unicode):
        path = path.encode('utf8')
    fixperm(path)

    header = environ.get('HTTP_RANGE')

    if not header:
        if xsend:
            headers.append(('X-Sendfile', path))
            return (200, None)
        else:
            return (200, _readfile(path))

    # ',' is not supported in Range
    ranged = re.findall('^bytes=(\d+)?-(\d+)?$', header)
    if ranged and ranged[0][0]:
        start = int(ranged[0][0])
        if ranged[0][1] and int(ranged[0][1]) < size:
            end = int(ranged[0][1])
        else: end = size - 1
    elif ranged and ranged[0][1]:
        start = size - int(ranged[0][1])
        end = size - 1
    elif ',' in header:
        raise StatusError(501)
    else:
        raise StatusError(416)

    content_range = 'bytes %s-%s/%s' % (start, end, size)
    headers.append(('Content-Range', content_range))
    if xsend:
        path = '%s %s-%s' % (urllib.quote(path), start, end)
        headers.append(('X-Sendfile2', path))
        body = None
    else:
        body = _readfile(path, start, end)

    return (206, body)

BUFFER = 2 * 1024 * 1024
def _write(_out, _in, count):
    while count:
        if count > BUFFER:
            size = BUFFER
        else:
            size = count
        count -= size
        try:
            os.write(_out, _in.read(size))
        except os.error as error:
            os.fsync(_out)
            fcntl.lockf(_out, fcntl.LOCK_UN)
            os.close(_out)
            if error.errno == errno.ENOSPC:
                raise StatusError(507)
            raise StatusError(503)
        except Exception as error:
            raise NASError('SystemError', error.message)

def _getboundary(line):
    try:
        enctype, subenctype = line.split(';')
        name, value = subenctype.split('=')
    except ValueError:
        return None
    if enctype.strip() == 'multipart/form-data' and name.strip() == 'boundary':
        return value.strip()
    return None

# '--boundary\r\n'
# 'Content-Disposition: ...\r\n'
# 'Content-Type: ...\r\n'
# '\r\n'
# FILE DATA
# '\r\n'
# '--boundary--\r\n'

def _skipboundary(boundary, post):
    size = 0
    inboundary = False
    while True:
        line = post.readline()
        size += len(line)
        if not line:
            break
        if line == '--%s\r\n' % boundary:
            inboundary = True
        elif line == '\r\n' and inboundary:
            break
    return size

def upload(path, offset=None, post=None, length=None, **kw):
    if post is None or length is None:
        raise StatusError(400)

    #if length == 0:
    #    raise StatusError(402)

    # errno  | None  |  0  | <size | >size
    #   -    | trunc | 204 |  204  |  416
    # ENOENT | 201   | 201 |  416  |  416
    # OTHER  | 503   | 503 |  503  |  503
    flags = os.O_WRONLY | os.O_TRUNC | os.O_SYNC
    try:
        sb = os.stat(path)
    except os.error as error:
        if error.errno != errno.ENOENT:
            raise StatusError(503)
        if offset:
            raise StatusError(416)
        status = 201
        _file = os.open(path, flags | os.O_CREAT, MODE)
    else:
        status = 204
        if offset is None:
            _file = os.open(path, flags, MODE)
        elif offset > sb.st_size:
            raise StatusError(416)
        else:
            _file = os.open(path, os.O_RDWR, MODE)
            os.lseek(_file, offset, os.SEEK_SET)

    # lock the file ...
    try:
        fcntl.lockf(_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError as error:
        os.close(_file)
        if error.errno in (errno.EAGAIN, errno.EACCES):
            # lockd
            raise StatusError(423)
        raise StatusError(503)

    # multipart/form-data; boundary=--
    boundary = _getboundary(kw.get('environ', {}).get('CONTENT_TYPE', ''))
    # we only suppose the easist form data
    if boundary is not None:
        length -= _skipboundary(boundary, post)
        length -= len('\r\n--%s--\r\n' % boundary)

    # write the file ...
    _write(_file, post, length)

    os.fsync(_file)
    fcntl.lockf(_file, fcntl.LOCK_UN)
    os.close(_file)
    return status

def _fixperm(path, login, mode):
    try:
        os.chmod(path, mode)
        pw = pwd.getpwnam(login)
        os.chown(path, pw.pw_uid, pw.pw_gid)
    except:
        pass

class StatusComplete(Exception):
    def __init__(self, status, headers=None, body=None):
        self.status = status

        self.headers = []
        if headers is None:
            headers = []

        hadlength = False
        for header in headers:
            if isinstance(header, (list, tuple)) and len(header) == 2:
                if header[0].lower() == 'transfer-encoding':
                    continue
                if header[0].lower() == 'content-length':
                    if hadlength:
                        continue
                    hadlength = True
                self.headers.append(tuple(header))

        self.body = []
        if isinstance(body, list):
            self.body.extend(body)
        elif isinstance(body, str):
            self.body.append(body)
            if not hadlength:
                self.headers.append(('Content-Length', str(len(body))))
        elif hasattr(body, 'getvalue'):
            self.body.append(body.getvalue())
            body.close()

def generate(path):
    _dir, _file = os.path.split(path)
    if _dir.endswith('.thumbnail') or _dir.endswith('.webview'):
        parent, thumb = os.path.split(_dir)
        realpath = os.path.join(parent, _file[:-4])
        if not os.path.exists(realpath):
            raise StatusError(404)
        if (os.path.exists(path) and os.path.getsize(path) != 0
                and os.path.getmtime(path) > os.path.getmtime(realpath)):
            return path
        result = convert.convert(realpath)
        path = result.get(thumb)
        if not path: raise StatusError(404)
    return path

def webdav_lockdiscovery(status, **kw):
    #locktoken, lockscope, locktype,
    #    depth, owner, timeout='Second-600'):

    headers = []
    headers.append(('Lock-Token', kw.get('locktoken')))
    headers.append(('Content-Type', 'text/xml; charset="utf-8"'))

    response = \
        '<?xml version="1.0" encoding="utf-8"?>' \
        '<D:prop xmlns:D="DAV:">' \
        '<D:lockdiscovery>' \
        '<D:activelock>' \
        '<D:locktype><D:%(locktype)s/></D:locktype>' \
        '<D:lockscope><D:%(lockscope)s/></D:lockscope>' \
        '<D:depth>%(depth)s</D:depth>' \
        '<D:owner>%(owner)s</D:owner>' \
        '<D:timeout>%(timeout)s</D:timeout>' \
        '<D:locktoken>' \
        '<D:href>%(locktoken)s</D:href>' \
        '</D:locktoken>' \
        '</D:activelock>' \
        '</D:lockdiscovery>' \
        '</D:prop>' % kw

    raise StatusComplete(status, headers, response)

class WEBDAV:
    # todo: authentication

    def OPTIONS(self, **kw):
        headers = []
        headers.append(('DAV', '1,2'))
        headers.append(('MS-Author-Via', 'DAV'))
        methods = ['GET', 'HEAD', 'PUT', 'DELETE',
                   'PROPFIND', 'MKCOL', 'COPY', 'MOVE',
                   'PROPPATCH', 'LOCK', 'UNLOCK']
        headers.append(('Allow', ', '.join(methods)))
        raise StatusComplete(200, headers)

    def GET(self, **kw):
        self.GETHEAD('GET', **kw)

    def HEAD(self, **kw):
        self.GETHEAD('HEAD', **kw)

    def GETHEAD(self, method, url, login=None, query='', **kw):
        # fixme
        if login is None:
            logins = urlparse.parse_qs(query, True).get('login')
            if logins:
                login = logins[0]

        download = False
        path, nil, write = url2path(url, login, query=query, kw=kw)
        if path == BACKUP:
            path = '/home/.nas/backup.zip'
            backup.backup(path)
            filename = 'isharing_backup_%s.zip' % time.strftime('%Y%m%d')
            download = True
        else:
            filename = path

        if path == REALFIRMWARE or path == REALRESTORE:
            raise StatusError(404)

        if path[-4:] == '.jpg' and write:
            path = generate(path)

        try:
            sb = os.stat(path)
        except os.error:
            raise StatusError(404)

        if stat.S_ISDIR(sb.st_mode):
            raise StatusError(403)
        elif not stat.S_ISREG(sb.st_mode):
            raise StatusError(404)

        path = os.path.realpath(path)
        if path[:8] != '/var/www' and method == 'GET' and kw.get('rotate', '') and write:
            try:
                rotate = int(kw.get('rotate'))
                im = Image.open(path)
                im.load()
                mi = im.rotate(rotate)
                mi.save(path)
            except:
                raise StatusError(304)

            parent, child = os.path.split(path)
            try:
                os.unlink(os.path.join(parent, '.thumbnail', '%s.jpg' % child))
            except os.error:
                pass
            try:
                os.unlink(os.path.join(parent, '.webview', '%s.jpg' % child))
            except os.error:
                pass
            if kw.get('nocontent') is not None:
                raise StatusError(204)

        checkif(sb, kw.get('environ', {}))

        status = 200
        headers = []
        body = None

        if kw.get('download', '').lower() == 'true':
            download = True

        if method == 'GET' and download:
            headers.append(('Content-Type', 'application/octet-stream'))
            headers.append(('Content-Disposition', _getdisposition(filename, **kw)))
        else:
            headers.append(('Content-Type', _getmimetype(path)))

        headers.append(('ETag', _getetag(sb)))
        headers.append(('Last-Modified', _getmtime(sb)))
        headers.append(('Accept-Ranges', 'bytes'))

        from info import _openvpn
        environ = kw.get('environ', {})
        if PLX7821 and _openvpn(environ) == 'yes':
            xsend = False
        else:
            xsend = True

        if method == 'HEAD':
            headers.append(('Content-Length', '%d' % sb.st_size))
        else: # GET
            status, body = _xsendfile(path, environ, sb.st_size, headers, xsend)

        if kw.get('exif') is not None:
            headers.extend(getexif(path))

        raise StatusComplete(status, headers, body)

    def PUT(self, **kw):
        self.POSTPUT(**kw)

    def POST(self, **kw):
        self.POSTPUT(**kw)

    def POSTPUT(self, url, login, query, **kw):
        path, nil, write = url2path(url, login, query=query)
        if not write:
            raise StatusError(403)
        if webdav_haslock(path, kw.get('environ', {})):
            raise StatusError(423)

        body = []
        headers = []

        try:
            sb = os.stat(path)
        except os.error:
            pass
        else:
            checkif(sb, kw.get('environ', {}))
            headers.append(('ETag', _getetag(sb)))
            headers.append(('Last-Modified', _getmtime(sb)))

        try:
            status = upload(path, **kw)
        except StatusError:
            pass
        except:
            raise StatusError(503)

        if kw.get('lastpackage') is not False:
            _fixperm(path, login, MODE)
            if path == REALFIRMWARE:
                try:
                    uncompress(path)
                except NASError:
                    status = 412
            elif path == REALRESTORE:
                try:
                    backup.restore(path)
                except NASError:
                    status = 412

        raise StatusComplete(status, headers, body)

    def DELETE(self, url, login, environ, query='', **kw):
        path, nil, write = url2path(url, login, query=query, kw=kw)
        if path == REALFIRMWARE:
            raise StatusError(404)

        if not write:
            raise StatusError(403)

        if webdav_haslock(path, environ):
            raise StatusError(423)

        try:
            mode = os.lstat(path).st_mode
        except os.error:
            raise StatusError(404)

        if stat.S_ISDIR(mode):
            try:
                shutil.rmtree(path)
            except (os.error, OSError):
                raise StatusError(503)
        else:
            try:
                os.unlink(path)
            except os.error as error:
                if error.errno == errno.ENOENT:
                    raise StatusError(404)
                raise StatusError(501)
        raise StatusComplete(204)

    def PROPFIND(self, url, login, environ, query='', **kw):

        depth = environ.get('HTTP_DEPTH', 'infinity')
        # we don't support Depth: infinity
        if depth == 'infinity':
            raise StatusError(501)
        if depth != '0' and depth != '1':
            raise StatusError(400)

        print 'url', url
        path, isdir, write = url2path(url, login, query, True)
        print 'path', path

        if depth == '1' and not isdir:
            raise StatusError(404)

        if path == REALFIRMWARE or path == REALRESTORE:
            raise StatusError(404)

        status = 207
        headers = []
        headers.append(('Content-Type', 'text/xml; charset="utf-8"'))

        try:
            sb = os.stat(path)
        except os.error:
            pass
        else:
            checkif(sb, kw.get('environ', {}))
            headers.append(('ETag', _getetag(sb)))
            headers.append(('Last-Modified', _getmtime(sb)))

        body = StringIO()
        body.write('<?xml version="1.0" encoding="utf-8"?>')
        body.write('<D:multistatus xmlns:D="DAV:" %s>' % DTNS)

        print 'getchilren', path
        items = getchildren(path, login, write, depth)

        for (child, write, filepath) in items:
            if write is not None:
                try:
                    st = os.stat(filepath)
                except os.error:
                    body.close()
                    raise StatusError(404)
                isdir = stat.S_ISDIR(st.st_mode)
                ctime = st.st_ctime
                mtime = st.st_mtime
                size = st.st_size
                mimetype = _getmimetype(filepath)
            else:
                isdir = True
                ctime = mtime = time.time()

            if child == '':
                newurl = url
            elif url[-1:] == '/':
                newurl = '%s%s' % (url, child)
            else:
                newurl = '%s/%s' % (url, child)
            if isinstance(newurl, unicode):
                newurl = newurl.encode('utf8')

            newurl = re.sub('/+', '/', newurl)

            body.write('<D:response>')
            body.write('<D:href>')
            body.write(urllib.quote(newurl))
            body.write('</D:href>')
            body.write('<D:propstat>')
            body.write('<D:prop>')

            # resourcetype
            if isdir:
                body.write('<D:resourcetype><D:collection/></D:resourcetype>')
                mimetype = 'httpd/unix-directory'
                size = 0

            # contenttype
            body.write('<D:getcontenttype>')
            body.write(mimetype)
            body.write('</D:getcontenttype>')

            # creationdate
            body.write('<D:creationdate dt:dt="dateTime.tz">')
            body.write(time.strftime(TZ, time.gmtime(ctime)))
            body.write('</D:creationdate>')

            # getlastmodified
            body.write('<D:getlastmodified dt:dt="dateTime.rfc1123">')
            body.write(time.strftime(RFC1123, time.gmtime(mtime)))
            body.write('</D:getlastmodified>')

            # getcontentlength
            body.write('<D:getcontentlength>')
            body.write('%d' % size)
            body.write('</D:getcontentlength>')

            # getcontentlanguage
            body.write('<D:getcontentlanguage>')
            body.write('en')
            body.write('</D:getcontentlanguage>')

            if write is not None and write is not True and write is not False:
                body.write('<R:write xmlns:R="works">')
                body.write(write and 'readwrite' or 'readonly')
                body.write('</R:write>')

            body.write('</D:prop>')
            body.write('<D:status>HTTP/1.1 200 OK</D:status>')
            body.write('</D:propstat>')
            body.write('</D:response>')

        body.write('</D:multistatus>')
        raise StatusComplete(status, headers, body)

    def MKCOL(self, url, login, query='', **kw):

        #if self.content_length != 0:
        #   raise StatusError(415)

        path, nil, write = url2path(url, login, query=query)

        if not write:
            raise StatusError(403)

        try: os.mkdir(path)
        except os.error as error:
            if error.errno == errno.EPERM:
                raise StatusError(403)
            elif error.errno == errno.ENOENT:
                raise StatusError(409)
            elif error.errno == errno.ENOTDIR:
                raise StatusError(409)
            elif error.errno == errno.EEXIST:
                raise StatusError(405)
            raise StatusError(405)
        _fixperm(path, login, 0775)
        raise StatusComplete(201)

    def COPY(self, url, login, environ, query='', **kw):
        self.COPYMOVE('COPY', url, login, environ, query)

    def MOVE(self, url, login, environ, query='', **kw):
        self.COPYMOVE('MOVE', url, login, environ, query)

    def COPYMOVE(self, method, url, login, environ, query):
        srcpath, isdir, srcwrite = url2path(url, login, query=query)

        # checklock for MOVE
        if method == 'MOVE' and webdav_haslock(srcpath, environ):
            raise StatusError(423)

        if not isdir and webdav_haslock(srcpath, environ):
            raise StatusError(423)

        header = environ.get('HTTP_DESTINATION', '')
        if not header:
            raise StatusError(400)

        destination = header
        result = urlparse.urlparse(destination)
        if result.hostname != environ.get('SERVER_NAME', ''):
            raise StatusError(502)
        dstpath, nil, dstwrite = url2path(result.path, login)

        if dstpath == srcpath or dstpath.startswith('%s/' % srcpath):
            raise StatusError(409)

        if not dstwrite:
            raise StatusError(403)

        if not srcwrite and method == 'MOVE':
            raise StatusError(403)

        header = environ.get('HTTP_OVERWRITE', 'T')
        if header != 'F' and header != 'T':
            raise StatusError(400)
        overwrite = header == 'T' and True or False

        header = environ.get('HTTP_DEPTH', 'infinity')
        if header != 'infinity' and header != '0':
            raise StatusError(400)
        depth = header == 'infinity' and True or False
        if method == 'MOVE' and not depth:
            raise StatusError(400)

        try:
            srcmode = os.stat(srcpath).st_mode
        except os.error as error:
            if error.errno == errno.ENOENT:
                raise StatusError(404)
            raise StatusError(403)
        if stat.S_ISDIR(srcmode):
            try:
                dstmode = os.stat(dstpath).st_mode
            except os.error:
                status = 201    # created
            else: # destination exists !
                status = 204    # no content
                if stat.S_ISDIR(dstmode):
                    if depth:
                        if not overwrite:
                            raise StatusError(412)
                        shutil.rmtree(dstpath)
                else:
                    if not overwrite:
                        raise StatusError(412)
                    os.unlink(dstpath)

            if method == 'MOVE':
                shutil.move(srcpath, dstpath)
            else: # COPY
                if depth:
                    shutil.copytree(srcpath, dstpath, True)
                else:
                    try:
                        mode = os.stat(dstpath).st_mode
                    except os.error:
                        os.mkdir(dstpath)
                    else:
                        if not stat.S_ISDIR(mode):
                            os.unlink(dstpath)
                            os.mkdir(dstpath)
                    shutil.copystat(srcpath, dstpath)
            raise StatusComplete(status)
        else: # ignore depth for normal file
            try:
                dstmode = os.stat(dstpath).st_mode
            except os.error:
                status = 201
            else:
                status = 204
                if stat.S_ISDIR(dstmode):
                    # not allow to replace file with directory
                    # it's too dangeraous !!!
                    raise StatusError(400)
                else:
                    if not overwrite:
                        raise StatusError(412)
                    os.unlink(dstpath)
            if method == 'MOVE':
                function = shutil.move
            else:
                function = shutil.copy

            try:
                function(srcpath, dstpath)
            except IOError as error:
                if error.errno == errno.ENOTDIR:
                    raise StatusError(424)
                elif errno.errno == errno.ENOENT:
                    raise StatusError(424)
                raise StatusError(405)
            raise StatusComplete(status)

    def LOCK(self, url, login, environ, post, length=0, query='', **kw):

        depth = environ.get('HTTP_DEPTH', 'infinity')
        if depth != '0' and depth != 'infinity':
            raise StatusError(400)

        path, isdir, write = url2path(url, login, query, True)
        if depth == 'infinity' and isdir:
            raise StatusError(409)

        def gettagname(xml, ns, tagname):
            node = xml.getElementsByTagNameNS(ns, tagname)[0]
            return node.firstChild.localName.encode('utf8')

        def gettagvalue(xml, ns, tagname, default=''):
            nodes = xml.getElementsByTagNameNS(ns, tagname)
            if nodes:
                return nodes[0].firstChild.toxml().encode('utf8')
            else:
                return default

        header = environ.get('HTTP_IF', '')
        if not header and length:
            try:
                xml = parseString(post.read(length))
                lockinfoxml = xml.getElementsByTagNameNS('DAV:', 'lockinfo')[0]
                lockscope = gettagname(lockinfoxml, 'DAV:', 'lockscope')
                locktype = gettagname(lockinfoxml, 'DAV:', 'locktype')
                owner = gettagvalue(lockinfoxml, 'DAV:', 'owner', login)
            except:
                raise StatusError(400)
            if lockscope != 'exclusive' and lockscope != 'shared':
                raise StatusError(400)
            if locktype != 'write':
                raise StatusError(400)

            _lockscope = _getlockscope(path)
            if _lockscope == 'exclusive':
                raise StatusError(423)
            if _lockscope and lockscope == 'exclusive':
                raise StatusError(423)
            locktoken = 'opaquelocktoken:%s' % uuid4()
            lockinfo = {'path': path,
                        'time': time.time(),
                        'status': 201,
                        'locktoken': locktoken,
                        'lockscope': lockscope,
                        'locktype': locktype,
                        'depth': depth,
                        'owner': owner,
                        'timeout': 'Second-600',}
            _addlockinfo(lockinfo)
            webdav_lockdiscovery(**lockinfo)
        else:
            index = 0
            lockscope = None
            while True:
                index = header.find('opaquelocktoken:', index)
                if index < 0:
                    break
                locktoken = header[index:index+52]
                lockscope, locktype, depth, owner = _refreshlockinfo(locktoken)
                if lockscope:
                    break
                index += 52
            if lockscope is None:
                raise StatusError(400)
            lockinfo = {'status': 200,
                        'locktoken': locktoken,
                        'lockscope': lockscope,
                        'locktype': locktype,
                        'depth': depth,
                        'owner': owner,
                        'timeout': 'Second-600',}
            webdav_lockdiscovery(**lockinfo)

    def UNLOCK(self, environ, **kw):
        header = environ.get('HTTP_LOCK_TOKEN', '')
        index = header.find('opaquelocktoken:')
        if index < 0:
            raise StatusError(400)
        locktoken = header[index:index+52]
        if _dellockinfo(locktoken):
            raise StatusError(204)
        else:
            raise StatusError(400)

    def PROPPATCH(self, url, post, length, **kw):

        if not length:
            raise StatusError(400)
        namespaces = {}
        props = []
        xmls = []
        try:
            xml = parseString(post.read(length))
            propxml = xml.getElementsByTagNameNS('DAV:', 'propertyupdate')[0]
            setxml = propxml.getElementsByTagNameNS('DAV:', 'set')
            removexml = propxml.getElementsByTagNameNS('DAV:', 'remove')
            for xml in setxml:
                xmls.append(xml)
            for xml in removexml:
                xmls.append(xml)
            for xml in xmls:
                prop = xml.getElementsByTagNameNS('DAV:', 'prop')[0]
                for child in prop.childNodes:
                    props.append('%s:%s' % (child.prefix, child.localName))
                    if child.prefix not in namespaces:
                        namespaces[child.prefix] = child.namespaceURI
        except:
            raise StatusError(400)
        headers = []
        headers.append(('Content-Type', 'text/xml; charset="utf-8"'))
        body = StringIO()
        body.write('<?xml version="1.0" encoding="utf-8"?>')
        body.write('<D:multistatus xmlns:D="DAV:"')
        for name, value in namespaces.iteritems():
            body.write(' xmlns:%s="%s"' % (str(name), str(value)))
        body.write('>')
        body.write('<D:response>')
        body.write('<D:href>%s</D:href>' % urllib.quote(url))
        for prop in props:
            body.write('<D:propstat>')
            body.write('<D:prop>')
            body.write('<%s/>' % str(prop))
            body.write('</D:prop>')
            body.write('<D:status>HTTP/1.1 200 OK</D:status>')
            body.write('</D:propstat>')
        body.write('</D:response>')
        body.write('</D:multistatus>')
        raise StatusComplete(207, headers, body)

LOCKINFO = []

def webdav_haslock(path, environ):
    header = environ.get('HTTP_IF', '')
    index = header.find('opaquelocktoken:')
    if index != -1:
        locktoken = header[index:index+52]
        lockscope, nil, nil, nil = _refreshlockinfo(locktoken)
        if lockscope != None:
            return False

    if _getlockscope(path):
        return True

    return False

def _getlockscope(path):
    # global LOCKINFO
    for lockinfo in LOCKINFO:
        if lockinfo.get('time') + 600 < time.time():
            LOCKINFO.remove(lockinfo)
            continue
        if lockinfo.get('path') == path:
            return lockinfo.get('lockscope')
    return False

def _addlockinfo(lockinfo):
    # global LOCKINFO
    LOCKINFO.append(lockinfo)

def _refreshlockinfo(locktoken):
    # global LOCKINFO
    for lockinfo in LOCKINFO:
        if lockinfo.get('locktoken') == locktoken:
            lockinfo['time'] = time.time()
            return (lockinfo['lockscope'],
                    lockinfo['locktype'],
                    lockinfo['depth'],
                    lockinfo['owner'])
    return (None, None, None, None)

def _dellockinfo(locktoken):
    # global LOCKINFO
    for lockinfo in LOCKINFO:
        if lockinfo.get('locktoken') == locktoken:
            LOCKINFO.remove(lockinfo)
            return True
    return False

responses = {
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',

    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi-Status',

    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    306: '(Unused)',
    307: 'Temporary Redirect',

    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed',
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',

    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported',
    507: 'Insufficient Storage',
}

__all__ = ['PathError', 'StatusError', 'StatusComplete', 'WEBDAV', 'responses', 'upload']

# vim: set sta sw=4 et:
