#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Tue 03 May 2011 01:02:17 PM CST

import os
import sys
import socket
import hashlib
import httplib
from system import NASError
from system import getrealpath
from webdav import StatusComplete

KEY = 'AI39si6VscapiINYSbxpW-BxcDUxFEu1St6QZRzy84BBYs6xn4IwN28Axj9ntay6OtuCkd2iuyRjzmWy3XcqiAfQ0TCQFxQNZQ'

BUFFER = 8192
HOST = 'uploads.gdata.youtube.com'
PATH = '/feeds/api/users/default/uploads'
XML = '''<?xml version="1.0"?>
<entry xmlns="http://www.w3.org/2005/Atom"
  xmlns:media="http://search.yahoo.com/mrss/"
  xmlns:yt="http://gdata.youtube.com/schemas/2007">
  <media:group>
    <media:title type="plain">%s</media:title>
    <media:category
      scheme="http://gdata.youtube.com/schemas/2007/categories.cat">%s
    </media:category>
    <yt:incomplete/>
  </media:group>
</entry>'''

def _postdata(token, filename, title=None, **kw):
    boundary = hashlib.sha1(os.urandom(20)).hexdigest()

    if title is None:
        title = os.path.basename(filename)

    conn = httplib.HTTPSConnection(HOST)
    conn.putrequest('POST', PATH)
    conn.putheader('Authorization', 'AuthSub token="%s"' % token)
    conn.putheader('GData-Version', '2')
    conn.putheader('X-GData-Key', 'key=%s' % KEY)
    conn.putheader('Slug', '%s' % title)

    conn.putheader('Content-Type', 'multipart/related; boundary="%s"' % boundary)

    prefix = ''
    suffix = ''

    prefix += '--%s\r\n' % boundary
    prefix += 'Content-Type: application/atom+xml; charset=UTF-8\r\n\r\n'
    prefix += XML % (title, kw.get('category', 'People'))

    prefix += '\r\n--%s\r\n' % boundary
    prefix += 'Content-Type: application/octet-stream\r\n'
    prefix += 'Content-Transfer-Encoding: binary\r\n\r\n'

    suffix += '\r\n--%s--\r\n' % boundary

    size = os.path.getsize(filename)
    content_length = len(prefix) + size + len(suffix)

    conn.putheader('Content-Length', '%d' % content_length)
    conn.putheader('Connection', 'Close')

    try:
        conn.endheaders()
    except socket.error as error:
        raise NASError('Proxy Error', error.strerror)

    conn.send(prefix)

    pipe = open(filename, 'rb')
    while size > 0:
        length = size > BUFFER and BUFFER or size
        data = pipe.read(length)
        if not data:
            break
        if len(data) != length:
            conn.close()
            pipe.close()
            raise NASError('Read Error')
        conn.send(data)
        size -= length
    pipe.close()

    conn.send(suffix)

    resp = conn.getresponse()
    status = resp.status
    headers = resp.getheaders()
    content = resp.read()
    resp.close()
    conn.close()

    raise StatusComplete(status, headers, content)

def getpath():
    import urllib

    SCOPE = 'http://gdata.youtube.com'
    path = 'https://www.google.com/accounts/AuthSubRequest'
    newquery = {'scope': SCOPE,
                'session': '1',
                'secure': '0',
                'next': 'http://thom.nklog.org/oauth/'}
    return '%s?%s' % (path, urllib.urlencode(newquery))

def upload(path, login=None, token=None, **kw):
    '''
    upload video to youtube.com

    arguments:
    path: the path in nas
    token: the token of authsub authentiacation
    title: optional, video title, default filename
    category: optional, video category, default People
    '''

    filename = getrealpath(path, login, False)
    if not filename:
        raise NASError('Not Found')

    if not token:
        raise NASError('No Token')

    try:
        _postdata(token, filename, **kw)
    except (NASError, StatusComplete):
        raise
    except Exception as error:
        if hasattr(error, 'strerror'):
            content = error.strerror
        else:
            content = ', '.join([str(x) for x in error.args])
        raise NASError('Proxy Error', content)

def proxy(path, token, host='gdata.youtube.com', **kw):
    '''
    proxy for youtube
    host: (optional), default 'gdata.youtube.com'
    path: path to youtube
    token: token

    Revoke Proxy:
    host = 'www.google.com'
    path = '/accounts/AuthSubSessionToken'
    '''
    conn = httplib.HTTPSConnection(host)
    conn.putrequest('GET', path)
    conn.putheader('Authorization', 'AuthSub token="%s"' % token)
    conn.putheader('GData-Version', '2')
    conn.putheader('X-GData-Key', 'key=%s' % KEY)
    conn.putheader('Content-Type', 'application/x-www-form-urlencoded')
    status = None
    headers = []
    content = ''
    try:
        conn.endheaders()
        resp = conn.getresponse()
        status = resp.status
        headers = resp.getheaders()
        content = resp.read()
        resp.close()
    except Exception as error:
        # make sure
        status = None
        if hasattr(error, 'strerror'):
            content = error.strerror
        else:
            content = ', '.join([str(x) for x in error.args])
    finally:
        conn.close()

    if status is None:
        raise NASError('Proxy Error', content)

    raise StatusComplete(status, headers, content)

__all__ = ['proxy', 'upload']

# vim: set sta sw=4 et:
